/*-------------------------------------------------------------------------
 *
 * int1.c
 *	  Functions for an 8-bit integer type
 *
 * Created by Randolf Richardson, 2017-Jul-20 through 2017-Jul-27
 * http://www.randolfrichardson.com/
 *
 * Portions Copyright (c) 2017, Randolf Richardson
 * Portions Copyright (c) 1996-2017, PostgreSQL Global Development Group
 * Portions Copyright (c) 1994, Regents of the University of California
 *
 * IDENTIFICATION
 *	  contrib/int1/int1.c
 *
 * VERSION HISTORY
 *        2017-Aug-02 v1.01 - Added support for GENERATE_SERIES
 *        2017-Jul-27 v1.00 - Initial version
 *
 *-------------------------------------------------------------------------
 */
#include "postgres.h"

#include <limits.h>
#include <math.h>

#include "fmgr.h"
#include "funcapi.h"
#include "access/hash.h"
#include "libpq/pqformat.h"
#include "utils/builtins.h"
#include "utils/numeric.h"
#include "utils/varbit.h"
#include "int1.h"

/*****************************************************************************
 * Intput/Output subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_in);
Datum int1_in(PG_FUNCTION_ARGS) {

    char *str = PG_GETARG_CSTRING(0);
    INT1   *b = palloc(sizeof(INT1));;

    b->b = (int8)pg_atoi(str, sizeof(b->b), '\0');

    PG_RETURN_INT1_P(b);
//    PG_RETURN_INT1(b->b); // Not working; use SELECT '1'::INT1 to test
}

PG_FUNCTION_INFO_V1(int1_out);
Datum int1_out(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);
    char *out = palloc(INT1_STRING_MAX);

    pg_itoa(b->b, out);

    PG_RETURN_CSTRING(out);
}

PG_FUNCTION_INFO_V1(int1_recv);
Datum int1_recv(PG_FUNCTION_ARGS) {

    StringInfo buf = (StringInfo)PG_GETARG_POINTER(0);
    INT1 *b = palloc(sizeof(INT1));

    b->b = (int8)pq_getmsgint(buf, sizeof(int8));

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_send);
Datum int1_send(PG_FUNCTION_ARGS) {

    INT1 *b = PG_GETARG_INT1_P(0);
    StringInfoData buf;

    pq_begintypsend(&buf);
    pq_sendint(&buf, b->b, sizeof(int8));
    PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

/*****************************************************************************
 * Cast subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_cast_from_bool);
Datum int1_cast_from_bool(PG_FUNCTION_ARGS) {

    bool  arg0 = PG_GETARG_BOOL(0);
    INT1    *b = palloc(sizeof(INT1));

    b->b = arg0 ? 1 : 0;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_bool);
Datum int1_cast_to_bool(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_BOOL(arg0->b != 0);
}

PG_FUNCTION_INFO_V1(int1_cast_from_bytea);
Datum int1_cast_from_bytea(PG_FUNCTION_ARGS) {

    bytea *arg0 = PG_GETARG_BYTEA_P(0);
    INT1     *b = palloc(sizeof(INT1));
    int32   len = VARSIZE(arg0) - VARHDRSZ;

    if (len == 0) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("bytea cannot be converted to an 8-bit integer")));
        PG_RETURN_NULL();
    } else if (len > 1) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("bytea value is out of range for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b->b = arg0->vl_dat[0];

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_bytea);
Datum int1_cast_to_bytea(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    StringInfoData buf;

    pq_begintypsend(&buf);
    pq_sendint(&buf, arg0->b, sizeof(int8));
    PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

PG_FUNCTION_INFO_V1(int1_cast_from_char);
Datum int1_cast_from_char(PG_FUNCTION_ARGS) {

    BpChar *arg0 = PG_GETARG_BPCHAR_P(0);//DatumGetBpCharP(0); //PG_GETARG_CHAR(0);
    INT1    *b = palloc(sizeof(INT1));

    b->b = (int8)arg0->vl_dat[0];

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_char);
Datum int1_cast_to_char(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);
    BpChar  *ch = palloc(sizeof(BpChar));

    SET_VARSIZE(ch, VARHDRSZ + 1);
    ch->vl_dat[0] = (int8)arg0->b;

    PG_RETURN_BPCHAR_P(ch);
}

PG_FUNCTION_INFO_V1(int1_cast_from_float4);
Datum int1_cast_from_float4(PG_FUNCTION_ARGS) {

    int32 arg0 = rint(PG_GETARG_FLOAT4(0));
    INT1    *b;

    if (arg0 > INT1_MAX || arg0 < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("float4 value is out of range for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0 & 0xff;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_float4);
Datum int1_cast_to_float4(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_FLOAT4((float4)arg0->b);
}

PG_FUNCTION_INFO_V1(int1_cast_from_float8);
Datum int1_cast_from_float8(PG_FUNCTION_ARGS) {

    int64 arg0 = rint(PG_GETARG_FLOAT8(0));
    INT1    *b;

    if (arg0 > INT1_MAX || arg0 < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("float8 value is out of range for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0 & 0xff;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_float8);
Datum int1_cast_to_float8(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_FLOAT8((float8)arg0->b);
}

PG_FUNCTION_INFO_V1(int1_cast_from_int2);
Datum int1_cast_from_int2(PG_FUNCTION_ARGS) {

    int16 arg0 = PG_GETARG_INT16(0);
    INT1    *b;

    if (arg0 > INT1_MAX || arg0 < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("int2 value is out of range for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0 & 0xff;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_int2);
Datum int1_cast_to_int2(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_INT16((int16)arg0->b);
}

PG_FUNCTION_INFO_V1(int1_cast_from_int4);
Datum int1_cast_from_int4(PG_FUNCTION_ARGS) {

    int32 arg0 = PG_GETARG_INT32(0);
    INT1    *b;

    if (arg0 > INT1_MAX || arg0 < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("int4 value %d is out of range for 8-bit integer", arg0)));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0 & 0xff;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_int4);
Datum int1_cast_to_int4(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_INT32((int32)arg0->b);
}

PG_FUNCTION_INFO_V1(int1_cast_from_int8);
Datum int1_cast_from_int8(PG_FUNCTION_ARGS) {

    int64 arg0 = PG_GETARG_INT64(0);
    INT1    *b;

    if (arg0 > INT1_MAX || arg0 < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("int8 value is out of range for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0 & 0xff;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_int8);
Datum int1_cast_to_int8(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_INT64((int64)arg0->b);
}

PG_FUNCTION_INFO_V1(int1_cast_from_numeric);
Datum int1_cast_from_numeric(PG_FUNCTION_ARGS) {

    Numeric      arg0 = PG_GETARG_NUMERIC(0);
    int32           v;
    INT1           *b;

    /* XXX would it be better to return NULL? */
    if (numeric_is_nan(arg0))
        ereport(ERROR,
            (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
             errmsg("cannot convert NaN to int1")));

    // Convert numeric to int4 (numeric_int2 uses more code than int4 or int8)
    v = DatumGetInt32(DirectFunctionCall1(numeric_int4, PointerGetDatum(arg0)));

    if (v > INT1_MAX || v < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("numeric value is out of range for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int16)v;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_numeric);
Datum int1_cast_to_numeric(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);

    PG_RETURN_NUMERIC(DirectFunctionCall1(int4_numeric, (int32)arg0->b));
}

PG_FUNCTION_INFO_V1(int1_cast_from_varbit);
Datum int1_cast_from_varbit(PG_FUNCTION_ARGS) {

    VarBit *arg0 = PG_GETARG_VARBIT_P(0);
    int32   bits = arg0->bit_len; // Number of bits
    INT1      *b;

    if (bits > 8) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("too many bits for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0->bit_dat[0] >> (8 - bits); // Bits are shifted left so we need to put them in their proper place

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_varbit);
Datum int1_cast_to_varbit(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);
    VarBit  *vb = palloc(9); // palloc(sizeof(VarBit));

    SET_VARSIZE(vb, VARBITHDRSZ + 5);
    vb->bit_len = 8; // 8 bits (1 byte)
    vb->bit_dat[0] = (int8)arg0->b;

    PG_RETURN_VARBIT_P(vb);
}

PG_FUNCTION_INFO_V1(int1_cast_from_bit);
Datum int1_cast_from_bit(PG_FUNCTION_ARGS) {

    VarBit *arg0 = PG_GETARG_VARBIT_P(0);
    int32   bits = arg0->bit_len; // Number of bits
    INT1      *b;

    if (bits > 8) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("too many bits for 8-bit integer")));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));

    b->b = (int8)arg0->bit_dat[0] >> (8 - bits); // Bits are shifted left so we need to put them in their proper place

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_cast_to_bit);
Datum int1_cast_to_bit(PG_FUNCTION_ARGS) {

    INT1  *arg0 = PG_GETARG_INT1_P(0);
    VarBit  *vb = palloc(9); // palloc(sizeof(VarBit)); // TODO:  Fix this

    SET_VARSIZE(vb, VARBITTOTALLEN(8));
    VARBITLEN(vb) = 8; // 8 bits (1 byte)
    vb->bit_dat[0] = (int8)arg0->b;

    PG_RETURN_VARBIT_P(vb);
}

/*****************************************************************************
 * Unary Operator subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_uminus);
Datum int1_uminus(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);

    b->b = -(b->b);
    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_uplus);
Datum int1_uplus(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_uflip);
Datum int1_uflip(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);

    b->b = ~(b->b);
    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_ureverse);
Datum int1_ureverse(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);
    int8    n = 0;
    int     i = 0;

    for (; i < 8; i++) {
      if ((b->b & (1 << i))) {
        n |= 1 << (7 - i);
      } // -x- if b->b -x-
    } // -x- for i -x-
    b->b = n;

    PG_RETURN_INT1_P(b);
}

PG_FUNCTION_INFO_V1(int1_abs);
Datum int1_abs(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);

    if (b->b < 0) {
       b->b = -b->b;
       /* overflow check (needed for INT_MIN) */
       if (b->b < 0)
            ereport(ERROR,
                (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
                 errmsg("int1 out of range")));
    }

    PG_RETURN_INT1_P(b);
}

/*****************************************************************************
 * Conversion Operator subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_cfrom_int4);
Datum int1_cfrom_int4(PG_FUNCTION_ARGS) {

    int32   arg0 = PG_GETARG_INT32(0);
    INT1      *b;

    if (arg0 > INT1_MAX || arg0 < INT1_MIN) {
        ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE), errmsg("int4 value %d is out of range for 8-bit integer", arg0)));
        PG_RETURN_NULL();
    }

    b = palloc(sizeof(INT1));
    b->b = (int8)arg0;

    PG_RETURN_INT1_P(b);
}

/*****************************************************************************
 * Comparison Operator subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_eq);
Datum int1_eq(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_BOOL(arg0->b == arg1->b);
}

PG_FUNCTION_INFO_V1(int1_ne);
Datum int1_ne(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_BOOL(arg0->b != arg1->b);
}

PG_FUNCTION_INFO_V1(int1_lt);
Datum int1_lt(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_BOOL(arg0->b < arg1->b);
}

PG_FUNCTION_INFO_V1(int1_le);
Datum int1_le(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_BOOL(arg0->b <= arg1->b);
}

PG_FUNCTION_INFO_V1(int1_gt);
Datum int1_gt(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_BOOL(arg0->b > arg1->b);
}

PG_FUNCTION_INFO_V1(int1_ge);
Datum int1_ge(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_BOOL(arg0->b >= arg1->b);
}

/*****************************************************************************
 * Bitwise Operator subroutines
 *
 * To do:  Reverse bit order int4_rev
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_and);
Datum int1_and(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    arg0->b &= arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_or);
Datum int1_or(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    arg0->b |= arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_xor);
Datum int1_xor(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    arg0->b ^= arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_shl);
Datum int1_shl(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    arg0->b <<= arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_shr);
Datum int1_shr(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    arg0->b >>= arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_shru);
Datum int1_shru(PG_FUNCTION_ARGS) {

    INT1   *arg0 = PG_GETARG_INT1_P(0);
    INT1   *arg1 = PG_GETARG_INT1_P(1);

    if (arg1->b > 0) arg0->b = (arg0->b >> 1) & 0x7f; // 1st bit:  Unsigned shift (mask to clear highest bit to convert 8-bit value to unsigned)
    if (arg1->b > 1) arg0->b >>= (arg1->b - 1); // Remaining bits:  Signed shift

    PG_RETURN_INT1_P(arg0);
}

/*****************************************************************************
 * Mathematical Operator subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_plus);
Datum int1_plus(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);
    int8     b = arg0->b + arg1->b;

    /*
     * Overflow check.  If the inputs are of different signs then their sum
     * cannot overflow.  If the inputs are of the same sign, their sum had
     * better be that sign too.
     */
    if (SAMESIGN(arg0, arg1) && !SAMESIGN(b, arg1))
        ereport(ERROR,
    	    (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
             errmsg("int1 out of range")));

    arg0->b = b;
    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_minus);
Datum int1_minus(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);
    int8     b = arg0->b - arg1->b;

    /*
     * Overflow check.  If the inputs are of different signs then their sum
     * cannot overflow.  If the inputs are of the same sign, their sum had
     * better be that sign too.
     */
    if (SAMESIGN(arg0, arg1) && !SAMESIGN(b, arg1))
        ereport(ERROR,
    	    (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
             errmsg("int1 out of range")));

    arg0->b = b;
    PG_RETURN_INT1(arg0);
}

PG_FUNCTION_INFO_V1(int1_mul);
Datum int1_mul(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);
    int32    b = arg0->b * arg1->b;

    /*
     * Overflow check.
     */
    if (b > INT1_MAX || b < INT1_MIN)
        ereport(ERROR,
                (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
                 errmsg("int1 out of range")));

    arg0->b = b;
    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_div);
Datum int1_div(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);

    /*
     * Division-by-Zero prevention.
     */
    if (arg1->b == 0) {
        ereport(ERROR,
            (errcode(ERRCODE_DIVISION_BY_ZERO),
             errmsg("int1 division by zero")));
        /* Ensure compiler realizes we mustn't reach the division by zero (GCC bug) */
        PG_RETURN_NULL();
    }

    /*
     * INT_MIN / -1 is problematic, since the result can't be represented on a
     * two's-complement machine.  Some machines produce INT_MIN, some produce
     * zero, some throw an exception.  We can dodge the problem by recognizing
     * that division by -1 is the same as negation.
     */
    if (arg1->b == -1)
    {
        arg1->b = -arg0->b;
        /* overflow check (needed for INT_MIN) */
        if (arg0->b != 0 && SAMESIGN(arg1->b, arg0->b))
            ereport(ERROR,
                (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
                 errmsg("int1 out of range")));
        PG_RETURN_INT1_P(arg1);
    }

    /* No overflow is possible */
    arg0->b = arg0->b / arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_mod);
Datum int1_mod(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);

    /*
     * Division-by-Zero prevention.
     */
    if (arg1->b == 0) {
        ereport(ERROR,
            (errcode(ERRCODE_DIVISION_BY_ZERO),
             errmsg("int1 division by zero")));
        /* Ensure compiler realizes we mustn't reach the division (gcc bug) */
        PG_RETURN_NULL();
    }

    /*
     * Some machines throw a floating-point exception for INT_MIN % -1, which
     * is a bit silly since the correct answer is perfectly well-defined,
     * namely zero.
     */
    if (arg1->b == -1)
    {
        arg0->b = 0;
        PG_RETURN_INT1_P(arg0);
    }

    /* No overflow is possible */
    arg0->b = arg0->b % arg1->b;

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_larger);
Datum int1_larger(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_INT1_P(arg0->b > arg1->b ? arg0 : arg1);
}

PG_FUNCTION_INFO_V1(int1_smaller);
Datum int1_smaller(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);

    PG_RETURN_INT1_P(arg0->b < arg1->b ? arg0 : arg1);
}

/*****************************************************************************
 * Utility subroutines
 *
 * Some of these subroutines are named in a manner consistent with other
 * well-known functions that process other data types (e.g., there are two
 * functions called "to_hex" that accept "int4" and "int8" types, and so it
 * is reasonable to assume that an SQL programmer will likely expect the same
 * subroutine name for this "int1" type).
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_inc);
Datum int1_inc(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);

    arg0->b++;

    /*
     * Overflow check.
     */
    if (arg0->b == INT1_MIN)
        ereport(ERROR,
    	    (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
             errmsg("int1 out of range")));

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_dec);
Datum int1_dec(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);

    arg0->b--;

    /*
     * Overflow check.
     */
    if (arg0->b == INT1_MAX)
        ereport(ERROR,
    	    (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
             errmsg("int1 out of range")));

    PG_RETURN_INT1_P(arg0);
}

PG_FUNCTION_INFO_V1(int1_hash);
Datum int1_hash(PG_FUNCTION_ARGS) {

    INT1   *b = PG_GETARG_INT1_P(0);
    int32 tmp = b->b & 0xff;

    return hash_uint32(tmp);
}

PG_FUNCTION_INFO_V1(to_hex);
Datum to_hex(PG_FUNCTION_ARGS) {

    INT1       *b = PG_GETARG_INT1_P(0);
    const char *DIGITS = "0123456789abcdef";
    int        i = 0;
    char       hex[3]; // One or two nybbles + ASCIIZ terminator

    // Fast conversion because we're only processing one or two nybbles
    if (b->b & 0xf0) hex[i++] = DIGITS[(b->b & 0xf0) >> 4]; // High nybble
    hex[i++] = DIGITS[(b->b & 0x0f)]; // Low nybble
    hex[i] = 0; // ASCIIZ terminator

    PG_RETURN_TEXT_P(cstring_to_text((char *)hex));
}

/*****************************************************************************
 * Generate Series subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(generate_series);
Datum generate_series(PG_FUNCTION_ARGS) {

    return generate_series_step(fcinfo);
}

PG_FUNCTION_INFO_V1(generate_series_step);
Datum generate_series_step(PG_FUNCTION_ARGS) {

    FuncCallContext              *funcctx;
    generate_series_int1_fctx    *fctx;
    MemoryContext                oldcontext;
    INT1                         *result = palloc(sizeof(INT1));

    if (SRF_IS_FIRSTCALL()) {

        INT1   *start = PG_GETARG_INT1_P(0);
        INT1  *finish = PG_GETARG_INT1_P(1);
        INT1    *step;

        // see if we were given an explicit step size
        if (PG_NARGS() == 3)
        {
            step = PG_GETARG_INT1_P(2);

            if (step->b == 0)
                ereport(ERROR,
                        (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
                         errmsg("int1 step size cannot equal zero")));
        } else {
            step = palloc(sizeof(INT1));
            step->b = 1; // Default is STEP 1
        }

        // create a function context for cross-call persistence
        funcctx = SRF_FIRSTCALL_INIT();

        //
        // Switch to memory context appropriate for multiple function calls.
        //
        oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);

        // allocate memory for user context
        fctx = (generate_series_int1_fctx *)palloc(sizeof(generate_series_int1_fctx));

        //
        // Use fctx to keep state from call to call. Seed current with the
        // original start value. We must copy the start_num and stop_num
        // values rather than pointing to them, since we may have detoasted
        // them in the per-call context.
        //
        fctx->current = start;
        fctx->finish  = finish;
        fctx->step    = step;

        funcctx->user_fctx = fctx;
        MemoryContextSwitchTo(oldcontext);
    }

    // stuff done on every call of the function
    funcctx = SRF_PERCALL_SETUP();

    //
    // Get the saved state and use current state as the result of this
    // iteration.
    //
    fctx      = funcctx->user_fctx;
    result->b = fctx->current->b; // Don't use result=fctx->current or the output will always be over-incremented/-decremented by 1 step

    if ((fctx->step->b > 0 && fctx->current->b <= fctx->finish->b) ||
        (fctx->step->b < 0 && fctx->current->b >= fctx->finish->b))
    {
        // increment current in preparation for next iteration
        fctx->current->b += fctx->step->b;

        // if next-value computation overflows, this is the final result
        if (SAMESIGN(result->b, fctx->step->b) && !SAMESIGN(result->b, fctx->current->b))
            fctx->step->b = 0;

        // do when there is more left to send
        SRF_RETURN_NEXT(funcctx, INT1_PGetDatum(result));
    }
    else {
        // do when there is no more left
        SRF_RETURN_DONE(funcctx);
    }
}

/*****************************************************************************
 * Index Support subroutines
 *****************************************************************************/

PG_FUNCTION_INFO_V1(int1_cmp);
Datum int1_cmp(PG_FUNCTION_ARGS) {

    INT1 *arg0 = PG_GETARG_INT1_P(0);
    INT1 *arg1 = PG_GETARG_INT1_P(1);
    int32    r;

    if (arg0->b < arg1->b) {
      r = -1;
    } else if (arg0->b == arg1->b) {
      r = 0;
    } else {
      r = 1;
    }

    PG_RETURN_INT32(r);
}
